
import traceback
from hrepr import StdHRepr
from weakref import WeakValueDictionary
from collections import deque
from types import FrameType


class UNAVAILABLE:
    def __init__(self):
        pass

    def __hrepr__(self, H, hrepr):
        return H.span('<UNAVAILABLE>', style='color:#bbb')


UNAVAILABLE = UNAVAILABLE()


class Registry:
    def __init__(self, keep=10000):
        self.keep = keep
        self.id = 0
        self.weak_map = WeakValueDictionary()
        self.strong_map = {}
        self.strong_ids = deque()

    def register(self, obj):
        self.id += 1
        try:
            self.weak_map[self.id] = obj
        except TypeError:
            self.strong_map[self.id] = obj
            self.strong_ids.append(self.id)
            if len(self.strong_ids) > self.keep >= 0:
                rm = self.strong_ids.popleft()
                del self.strong_map[rm]
        return self.id

    def resolve(self, id):
        try:
            return self.weak_map[id]
        except KeyError:
            return self.strong_map.get(id, UNAVAILABLE)


id_registry = Registry()


def handle_exception(e, H, hrepr):
    tb = e.__traceback__
    entries = traceback.extract_tb(tb)
    first = e.args[0] if len(e.args) > 0 else None
    iss = isinstance(first, str)
    args_table = H.table()
    for i in range(1 if iss else 0, len(e.args)):
        arg = e.args[i]
        tr = H.tr(H.td(i), H.td(hrepr(arg)))
        args_table = args_table(tr)

    views = H.tabbedView['hrepr-Exception'](
        H.strong(type(e).__name__),
        f': {first}' if iss else '',
        args_table
    )
    if not entries:
        return views

    last = entries[-1]
    for entry in entries:
        filename, lineno, funcname, line = entry
        absfile = os.path.abspath(filename)
        view = H.view()
        tab = H.tab(
            f'{funcname}@{os.path.basename(absfile)}'
        )
        snippet = H.codeSnippet(
            src = absfile,
            language = "python",
            line = lineno,
            context = 4
        )
        if filename.startswith('<'):
            snippet = snippet(getattr(e, '__repl_string__', ""))

        pane = H.pane(snippet)
        if entry is last:
            view = view(active = True)
        views = views(view(tab, pane))

    return views


def handle_frame(fr, H, hrepr):
    views = H.tabbedView()
    code = fr.f_code
    info = dict(
        name=code.co_name,
        filename=code.co_filename,
        firstlineno=code.co_firstlineno,
        lineno=fr.f_lineno
    )
    snippet = H.codeSnippet(
        src = code.co_filename,
        language = "python",
        line = fr.f_lineno,
        context = 4
    )
    views = views(H.view(H.tab('code'), H.pane(snippet)))
    views = views(H.view(H.tab('info'), H.pane(hrepr(info))))
    return views


class HRepr(StdHRepr):
    def __default_handlers__(self):
        h = super().__default_handlers__()
        h.update({
            Exception: handle_exception,
            FrameType: handle_frame
        })
        return h

    def __call__(self, obj, **kwargs):
        res = super().__call__(obj, **kwargs)
        try:
            the_id = id_registry.register(obj)
        except TypeError as e:
            return res
        else:
            return self.H.replObject({'obj-id': the_id})(res)




import sys
from .buche import *
from .repr import *
from .event import *


def _print_flush(x):
    print(x, flush=True)


master = MasterBuche(HRepr(), _print_flush)
buche = Buche(master, '/')
reader = Reader(sys.stdin)
read = reader.read




import json
import re
from threading import Thread
from .event import EventDispatcher
from .repr import id_registry


class ObjDict:
    def __init__(self, d):
        self.__dict__.update(d)

    def __hrepr__(self, H, hrepr):
        return hrepr(self.__dict__)


class BucheSeq:
    def __init__(self, objects):
        self.objects = list(objects)

    @classmethod
    def __hrepr_resources__(cls, H):
        return H.style('''
        .multi-print { display: flex; align-items: center; }
        .multi-print > * { margin-right: 10px; }
        ''')

    def __hrepr__(self, H, hrepr):
        return H.div['multi-print'](*map(hrepr, self.objects))


class MasterBuche:
    def __init__(self, hrepr, write):
        self.hrepr = hrepr
        self.write = write
        self.resources = set()

    def send(self, d={}, **params):
        message = {**d, **params}
        if 'path' not in message:
            raise ValueError(f'Must specify path for message: {message}')
        if 'command' not in message:
            raise ValueError(f'Must specify command for message: {message}')
        self.write(json.dumps(message))

    def show(self, obj, hrepr_params={}, **params):
        x = self.hrepr(obj, **hrepr_params)
        for res in self.hrepr.resources - self.resources:
            self.send(
                command = 'resource',
                path = '/',
                type = 'direct',
                contents = str(res)
            )
            self.resources.add(res)
        self.send(command='log',
                  format='html',
                  contents=str(x),
                  **params)


class Buche:
    def __init__(self, master, path, type=None, params=None, opened=False):
        self.master = master
        self.path = path
        self.type = type
        self.params = params or {}
        self.opened = opened
        self.subchannels = {}

    def _open(self):
        if self.type is not None:
            self.master.send(command='open', path=self.path,
                             type=self.type, **self.params)
        self.opened = True

    def configure(self, type=None, **params):
        if self.opened:
            raise Exception('Cannot configure a buche channel after it has'
                            'been opened.')
        if type:
            self.type = type
        self.params.update(params)

    def send(self, path=None, **params):
        if not self.opened:
            self._open()
        if path is None:
            path = self.path
        elif path.startswith('/'):
            pass
        else:
            path = self.join_path(path)
        self.master.send(path=path, **params)

    def log(self, contents, format='text', **params):
        self.send(command='log', format=format,
                   content=contents, **params)

    def pre(self, contents, **params):
        self.log(contents, format = 'pre', **params)

    def html(self, contents, **params):
        self.log(contents, format = 'html', **params)

    def markdown(self, contents, **params):
        self.log(contents, format = 'markdown', **params)

    def open(self, name, type, params, force=False):
        if not self.opened:
            self._open()
        if name.startswith('/'):
            raise ValueError('Channel name for open() cannot start with /')
        parts = re.split('/+', name, 1)
        if len(parts) == 1:
            if name in self.subchannels:
                return self.subchannels[name]
            else:
                ch = Buche(self.master, self.join_path(name), type, params)
                if force:
                    ch._open()
                self.subchannels[name] = ch
                return ch
        else:
            name, rest = parts
            return self.open(name, 'tabs', {}).open(rest, type, params, force)

    def __getattr__(self, attr):
        if attr.startswith('log_'):
            command = attr[4:]
            def _log(contents=None, **params):
                self.send(command=command, contents=contents, **params)
            return _log
        elif attr.startswith('open_'):
            chtype = attr[5:]
            def _open(name, **params):
                return self.open(name, chtype, params)
            return _open
        elif attr.startswith('ch_'):
            chname = attr[3:]
            return self[chname]
        else:
            raise AttributeError(f"'Buche' object has no attribute '{attr}'")

    def join_path(self, p):
        return f'{self.path.rstrip("/")}/{p.strip("/")}'

    def __getitem__(self, item):
        descr = item.split('::')
        if len(descr) == 1:
            type = None
            name, = descr
        else:
            name, type = descr
        return self.open(name, type, {})

    def show(self, obj, **params):
        self.master.show(obj, path=self.path, **params)

    def __call__(self, *objs, **keys):
        if len(objs) == 1 and not keys:
            o, = objs
        else:
            o = BucheSeq(objs)
            if keys:
                o.objects.append(keys)
        self.show(o)


class Reader(EventDispatcher):
    def __init__(self, source):
        super().__init__()
        self.source = source
        self.thread = Thread(target=self.loop)

    def read(self):
        line = self.source.readline()
        return self.parse(line)

    def parse(self, line):
        d = json.loads(line)
        if 'objId' in d and 'obj' not in d:
            d['obj'] = id_registry.resolve(int(d['objId']))
        message = ObjDict(d)
        self.emit(d.get('command', 'UNDEFINED'), message)
        return message

    def __iter__(self):
        for line in self.source:
            yield self.parse(line)

    def loop(self):
        for _ in self:
            pass

    def start(self):
        self.thread.start()
